# -*- coding: utf-8 -*-

from __future__ import absolute_import

__all__ = [
    'TSocketBase', 'TSocket', 'TServerSocket',
    'TTransportBase', 'TTransportException',
    'TMemoryBuffer',
    'TFramedTransport', 'TFramedTransportFactory',
    'TBufferedTransport', 'TBufferedTransportFactory',
    'TCyBufferedTransport', 'TCyBufferedTransportFactory',
]


from .socket import TSocketBase, TSocket, TServerSocket
from .transport import (
    TTransportBase,
    TTransportException,
    TMemoryBuffer,
    TBufferedTransport,
    TBufferedTransportFactory,
    TFramedTransport,
    TFramedTransportFactory,
)
from .cytransport import TCyBufferedTransport, TCyBufferedTransportFactory



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import struct

from io import BytesIO

from ..thrift import TType, TException


class TTransportBase(object):
    """Base class for Thrift transport layer."""

    def read(self, sz):
        buff = b''
        have = 0
        while have < sz:
            chunk = self._read(sz - have)
            have += len(chunk)
            buff += chunk

            if len(chunk) == 0:
                raise TTransportException(TTransportException.END_OF_FILE,
                                          "End of file reading from transport")

        return buff


class TTransportException(TException):
    """Custom Transport Exception class"""

    thrift_spec = {
        1: (TType.STRING, 'message'),
        2: (TType.I32, 'type'),
    }

    UNKNOWN = 0
    NOT_OPEN = 1
    ALREADY_OPEN = 2
    TIMED_OUT = 3
    END_OF_FILE = 4

    def __init__(self, type=UNKNOWN, message=None):
        super(TTransportException, self).__init__()
        self.type = type
        self.message = message


class TMemoryBuffer(TTransportBase):
    """Wraps a BytesIO object as a TTransport."""

    def __init__(self, value=None):
        """value -- a value as the initial value in the BytesIO object.

        If value is set, the transport can be read first.
        """
        self._buffer = BytesIO(value) if value is not None else BytesIO()
        self._pos = 0

    def isOpen(self):
        return not self._buffer.closed

    def open(self):
        pass

    def close(self):
        self._buffer.close()

    def read(self, sz):
        return self._read(sz)

    def _read(self, sz):
        orig_pos = self._buffer.tell()
        self._buffer.seek(self._pos)
        res = self._buffer.read(sz)
        self._buffer.seek(orig_pos)
        self._pos += len(res)
        return res

    def write(self, buf):
        self._buffer.write(buf)

    def flush(self):
        pass

    def getvalue(self):
        return self._buffer.getvalue()

    def setvalue(self, value):
        self._buffer = BytesIO(value)
        self._pos = 0


class TBufferedTransport(TTransportBase):
    """Class that wraps another transport and buffers its I/O.

    The implementation uses a (configurable) fixed-size read buffer
    but buffers all writes until a flush is performed.
    """
    DEFAULT_BUFFER = 4096

    def __init__(self, trans, rbuf_size=DEFAULT_BUFFER):
        self.__trans = trans
        self.__wbuf = BytesIO()
        self.__rbuf = BytesIO(b"")
        self.__rbuf_size = rbuf_size

    def is_open(self):
        return self.__trans.is_open()

    def open(self):
        return self.__trans.open()

    def close(self):
        return self.__trans.close()

    def _read(self, sz):
        ret = self.__rbuf.read(sz)
        if len(ret) != 0:
            return ret

        self.__rbuf = BytesIO(self.__trans.read(max(sz, self.__rbuf_size)))
        return self.__rbuf.read(sz)

    def write(self, buf):
        self.__wbuf.write(buf)

    def flush(self):
        out = self.__wbuf.getvalue()
        # reset wbuf before write/flush to preserve state on underlying failure
        self.__wbuf = BytesIO()
        self.__trans.write(out)
        self.__trans.flush()


class TBufferedTransportFactory(object):
    def get_transport(self, trans):
        return TBufferedTransport(trans)


class TFramedTransport(TTransportBase):
    """Class that wraps another transport and frames its I/O when writing."""
    def __init__(self, trans):
        self.__trans = trans
        self.__rbuf = BytesIO()
        self.__wbuf = BytesIO()

    def isOpen(self):
        return self.__trans.isOpen()

    def open(self):
        return self.__trans.open()

    def close(self):
        return self.__trans.close()

    def read(self, sz):
        ret = self.__rbuf.read(sz)
        if len(ret) != 0:
            return ret

        self.readFrame()
        return self.__rbuf.read(sz)

    def readFrame(self):
        buff = self.__trans.read(4)
        sz, = struct.unpack('!i', buff)
        self.__rbuf = BytesIO(self.__trans.read(sz))

    def write(self, buf):
        self.__wbuf.write(buf)

    def flush(self):
        wout = self.__wbuf.getvalue()
        wsz = len(wout)
        # reset wbuf before write/flush to preserve state on underlying failure
        self.__wbuf = BytesIO()
        # N.B.: Doing this string concatenation is WAY cheaper than making
        # two separate calls to the underlying socket object. Socket writes in
        # Python turn out to be REALLY expensive, but it seems to do a pretty
        # good job of managing string buffer operations without excessive
        # copies
        buf = struct.pack("!i", wsz) + wout
        self.__trans.write(buf)
        self.__trans.flush()


class TFramedTransportFactory(object):
    def get_transport(self, trans):
        return TFramedTransport(trans)



from libc.stdlib cimport malloc, free
from libc.string cimport memcpy, memmove

DEF DEFAULT_BUFFER = 4096
DEF MIN_BUFFER_SZIE = 1024


cdef class TCyBuffer(object):
    def __cinit__(self, buf_size):
        self.buf = <char*>malloc(buf_size)
        self.buf_size = buf_size
        self.cur = 0
        self.data_size = 0

    def __dealloc__(self):
        if self.buf != NULL:
            free(self.buf)
            self.buf = NULL

    cdef void move_to_start(self):
        memmove(self.buf, self.buf + self.cur, self.data_size)
        self.cur = 0

    cdef void clean(self):
        self.cur = 0
        self.data_size = 0

    cdef int grow(self, int min_size):
        if min_size <= self.buf_size:
            return 0

        cdef int multiples = min_size / self.buf_size
        if min_size % self.buf_size != 0:
            multiples += 1

        cdef int new_size = self.buf_size * multiples
        cdef char *new_buf = <char*>malloc(new_size)
        if new_buf == NULL:
            return -1
        memcpy(new_buf + self.cur, self.buf + self.cur, self.data_size)
        free(self.buf)
        self.buf_size = new_size
        self.buf = new_buf
        return 0


cdef class TCyBufferedTransport(object):
    """binary reader/writer"""

    def __init__(self, trans, int buf_size=DEFAULT_BUFFER):
        if buf_size < MIN_BUFFER_SZIE:
            raise Exception("buffer too small")

        self.trans = trans
        self.rbuf = TCyBuffer(buf_size)
        self.wbuf = TCyBuffer(buf_size)

    def is_open(self):
        return self.trans.is_open()

    def open(self):
        return self.trans.open()

    def close(self):
        return self.trans.close()

    def write(self, bytes data):
        cdef int sz = len(data)
        return self.c_write(data, sz)

    def read(self, int sz):
        cdef char* out = <char*>malloc(sz)
        self.c_read(sz, out)
        try:
            return out[:sz]
        finally:
            free(out)

    def flush(self):
        return self.c_flush()

    cdef c_write(self, char* data, int sz):
        cdef int cap = self.wbuf.buf_size - self.wbuf.data_size

        if cap < sz:
            if sz > self.wbuf.buf_size:
                if self.wbuf.grow(sz) != 0:
                    raise MemoryError("grow write buffer fail")
            self.c_flush()

        memcpy(self.wbuf.buf + self.wbuf.data_size, data, sz)
        self.wbuf.data_size += sz

    cdef c_read(self, int sz, char* out):
        if sz > self.rbuf.buf_size:
            if self.rbuf.grow(sz) != 0:
                raise MemoryError("grow read buffer fail")

        cdef int cap, new_data_len
        cdef bytes new_data, ret_str

        if self.rbuf.data_size < sz:
            cap = self.rbuf.buf_size - self.rbuf.data_size
            new_data = self.trans.read(cap)
            new_data_len = len(new_data)

            if new_data_len < sz - self.rbuf.data_size:
                raise Exception("Transport error")

            if cap - self.rbuf.cur < new_data_len:
                self.rbuf.move_to_start()

            memcpy(self.rbuf.buf + self.rbuf.cur + self.rbuf.data_size,
                   <char*>new_data, new_data_len)
            self.rbuf.data_size += new_data_len

        memcpy(out, self.rbuf.buf + self.rbuf.cur, sz)

        self.rbuf.cur += sz
        self.rbuf.data_size -= sz

    cdef c_flush(self):
        cdef bytes data
        if self.wbuf.data_size > 0:
            data = self.wbuf.buf[:self.wbuf.data_size]
            self.trans.write(data)
            self.wbuf.clean()

    def getvalue(self):
        return self.trans.getvalue()


class TCyBufferedTransportFactory(object):
    def get_transport(self, trans):
        return TCyBufferedTransport(trans)



# -*- coding: utf-8 -*-

__version__ = '0.1.10'

import sys
__python__ = sys.version_info

__all__ = ["install_import_hook", "remove_import_hook",  "load", "load_module"]

from .hook import install_import_hook, remove_import_hook
from .parser import load, load_module



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import errno
import os
import socket
import sys

from .transport import TTransportBase, TTransportException


class TSocketBase(TTransportBase):
    def _resolveAddr(self):
        if self._unix_socket is not None:
            return [(socket.AF_UNIX, socket.SOCK_STREAM, None, None,
                     self._unix_socket)]
        else:
            return socket.getaddrinfo(
                self.host,
                self.port,
                socket.AF_UNSPEC,
                socket.SOCK_STREAM,
                0,
                socket.AI_PASSIVE | socket.AI_ADDRCONFIG)

    def close(self):
        if self.handle:
            self.handle.close()
            self.handle = None


class TSocket(TSocketBase):
    """Socket implementation of TTransport base."""

    def __init__(self, host='localhost', port=9090, unix_socket=None):
        """Initialize a TSocket

        @param host(str)    The host to connect to.
        @param port(int)    The (TCP) port to connect to.
        @param unix_socket(str)  The filename of a unix socket to connect to.
                                 (host and port will be ignored.)
        """
        self.host = host
        self.port = port
        self.handle = None
        self._unix_socket = unix_socket
        self._timeout = None

    def set_handle(self, h):
        self.handle = h

    def is_open(self):
        return bool(self.handle)

    def set_timeout(self, ms):
        self._timeout = ms / 1000.0 if ms else None

        if self.handle:
            self.handle.settimeout(self._timeout)

    def open(self):
        try:
            res0 = self._resolveAddr()
            for res in res0:
                self.handle = socket.socket(res[0], res[1])
                self.handle.settimeout(self._timeout)
                try:
                    self.handle.connect(res[4])
                except socket.error as e:
                    if res is not res0[-1]:
                        continue
                    else:
                        raise e
                break
        except socket.error as e:
            if self._unix_socket:
                message = 'Could not connect to socket %s' % self._unix_socket
            else:
                message = 'Could not connect to %s:%d' % (self.host, self.port)
            raise TTransportException(type=TTransportException.NOT_OPEN,
                                      message=message)

    def read(self, sz):
        try:
            buff = self.handle.recv(sz)
        except socket.error as e:
            if (e.args[0] == errno.ECONNRESET and
                    (sys.platform == 'darwin' or
                     sys.platform.startswith('freebsd'))):
                # freebsd and Mach don't follow POSIX semantic of recv
                # and fail with ECONNRESET if peer performed shutdown.
                # See corresponding comment and code in TSocket::read()
                # in lib/cpp/src/transport/TSocket.cpp.
                self.close()
                # Trigger the check to raise the END_OF_FILE exception below.
                buff = ''
            else:
                raise
        if len(buff) == 0:
            raise TTransportException(type=TTransportException.END_OF_FILE,
                                      message='TSocket read 0 bytes')
        return buff

    def write(self, buff):
        if not self.handle:
            raise TTransportException(type=TTransportException.NOT_OPEN,
                                      message='Transport not open')
        sent = 0
        have = len(buff)
        while sent < have:
            plus = self.handle.send(buff)
            if plus == 0:
                raise TTransportException(type=TTransportException.END_OF_FILE,
                                          message='TSocket sent 0 bytes')
            sent += plus
            buff = buff[plus:]

    def flush(self):
        pass


class TServerSocket(TSocketBase):
    """Socket implementation of TServerTransport base."""

    def __init__(self, host=None, port=9090, unix_socket=None,
                 socket_family=socket.AF_UNSPEC):
        self.host = host
        self.port = port
        self._unix_socket = unix_socket
        self._socket_family = socket_family
        self.handle = None

    def listen(self):
        res0 = self._resolveAddr()
        socket_family = self._socket_family == socket.AF_UNSPEC and (
            socket.AF_INET6 or self._socket_family)
        for res in res0:
            if res[0] is socket_family or res is res0[-1]:
                break

        # We need remove the old unix socket if the file exists and
        # nobody is listening on it.
        if self._unix_socket:
            tmp = socket.socket(res[0], res[1])
            try:
                tmp.connect(res[4])
            except socket.error as err:
                eno, message = err.args
                if eno == errno.ECONNREFUSED:
                    os.unlink(res[4])

        self.handle = socket.socket(res[0], res[1])
        self.handle.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if hasattr(self.handle, 'settimeout'):
            self.handle.settimeout(None)
        self.handle.bind(res[4])
        self.handle.listen(128)

    def accept(self):
        client, addr = self.handle.accept()
        result = TSocket()
        result.set_handle(client)
        return result



cdef class TCyBuffer(object):
    cdef char *buf
    cdef int cur, buf_size, data_size

    cdef void move_to_start(self)
    cdef void clean(self)
    cdef int grow(self, int min_size)


cdef class TCyBufferedTransport(object):
    cdef object trans
    cdef TCyBuffer rbuf, wbuf

    cdef c_read(self, int sz, char* out)
    cdef c_write(self, char* data, int sz)
    cdef c_flush(self)



# -*- coding: utf-8 -*-

"""
    thriftpy.parser
    ~~~~~~~~~~~~~~~

    Thrift parser using ply
"""

from __future__ import absolute_import

import itertools
import os
import sys
import types

from .parser import parse
from ..thrift import gen_init, TException, TPayload, TType


_thriftloader = {}


def load(thrift_file, module_name=None):
    """Load thrift_file as a module

    The module loaded and objects inside may only be pickled if module_name
    was provided.
    """
    if module_name and not module_name.endswith("_thrift"):
        raise ValueError(
            "ThriftPy can only generate module with '_thrift' suffix")

    global _thriftloader
    # if module_name provided, it should be unique. we'll ignore the filename
    # here for the global thriftloader
    _thriftloader_key = module_name or thrift_file
    if _thriftloader_key in _thriftloader:
        return _thriftloader[_thriftloader_key]

    real_module = bool(module_name)

    # generate a fake module_name when it's not provided
    if not module_name:
        basename = os.path.basename(thrift_file)
        module_name, _ = os.path.splitext(basename)

    with open(thrift_file, "r") as fp:
        schema = fp.read()

    result = parse(schema)
    struct_names = list(itertools.chain(result["structs"].keys(),
                                        result["unions"].keys(),
                                        result["exceptions"].keys()))

    # load thrift schema as module
    thrift_schema = types.ModuleType(module_name)
    _type = lambda n, o: type(n, (o, ), {"__module__": module_name})

    def _ttype(t):
        if isinstance(t, str):
            if t in struct_names:
                return TType.STRUCT, getattr(thrift_schema, t)
            elif t in result["enums"]:
                return TType.I32
            elif t in result["typedefs"]:
                return _ttype(result["typedefs"][t])
            else:
                return getattr(TType, t.upper())

        if t[0] == "list":
            return TType.LIST, _ttype(t[1])
        elif t[0] == "map":
            return TType.MAP, (_ttype(t[1]), _ttype(t[2]))
        else:
            raise Exception("ttype parse error: {0}".format(t))

    def _ttype_spec(ttype, name):
        ttype = _ttype(ttype)
        if isinstance(ttype, int):
            return ttype, name
        else:
            return ttype[0], name, ttype[1]

    # load consts
    for name, value in result["consts"].items():
        setattr(thrift_schema, name, value)

    # load enums
    for name, enum in result["enums"].items():
        attrs = {"__module__": module_name}
        v2n, n2v = {}, {}
        for key, value in enum.items():
            attrs[key] = value
            v2n[value] = key
            n2v[key] = value
        attrs['_VALUES_TO_NAMES'] = v2n
        attrs['_NAMES_TO_VALUES'] = n2v
        enum_cls = type(name, (object, ), attrs)
        setattr(thrift_schema, name, enum_cls)

    # load structs/unions
    for name in itertools.chain(result["structs"].keys(),
                                result["unions"].keys()):
        attrs = {"__module__": module_name}
        struct_cls = type(name, (TPayload, ), attrs)
        setattr(thrift_schema, name, struct_cls)

    for name in result["exceptions"].keys():
        attrs = {"__module__": module_name}
        exc_cls = type(name, (TException, ), attrs)
        setattr(thrift_schema, name, exc_cls)

    for name, struct in itertools.chain(result["structs"].items(),
                                        result["unions"].items(),
                                        result["exceptions"].items()):
        thrift_spec, default_spec = {}, []
        for m in struct:
            thrift_spec[m["id"]] = _ttype_spec(m["type"], m["name"])
            default_spec.append((m["name"], m["value"] or None))
        struct_cls = getattr(thrift_schema, name)
        gen_init(struct_cls, thrift_spec, sorted(default_spec))

    # load services
    for name, service in result["services"].items():
        service_cls = _type(name, object)
        thrift_services = []
        for api_name, api in service["apis"].items():
            thrift_services.append(api_name)

            # generate default spec from thrift spec
            _default_spec = lambda s: [(s[k][1], None) for k in sorted(s)]

            # api args payload
            args_name = "%s_args" % api_name
            args_attrs = {"__module__": module_name}

            args_thrift_spec = {}
            for field in api["fields"]:
                args_thrift_spec[field["id"]] = _ttype_spec(field["type"],
                                                            field["name"])
            args_cls = type(args_name, (TPayload, ), args_attrs)
            gen_init(args_cls, args_thrift_spec,
                     _default_spec(args_thrift_spec))
            setattr(service_cls, args_name, args_cls)

            # api result payload
            result_name = "%s_result" % api_name
            result_attrs = {"__module__": module_name}

            if api["type"] == "void":
                result_thrift_spec = {}
            else:
                result_thrift_spec = {0: _ttype_spec(api["type"], "success")}

            if api.get("throws"):
                for t in api["throws"]:
                    result_thrift_spec[t["id"]] = _ttype_spec(t["type"],
                                                              t["name"])

            result_cls = type(result_name, (TPayload, ), result_attrs)
            gen_init(result_cls, result_thrift_spec,
                     _default_spec(result_thrift_spec))
            setattr(service_cls, result_name, result_cls)

        setattr(service_cls, "thrift_services", thrift_services)
        setattr(thrift_schema, name, service_cls)
    thrift_schema.__file__ = thrift_file

    _thriftloader[_thriftloader_key] = thrift_schema
    # insert into sys.modules if module_name was provided manually
    if real_module:
        sys.modules[module_name] = _thriftloader[_thriftloader_key]
    return _thriftloader[_thriftloader_key]


def _import_module(import_name):
    if '.' in import_name:
        module, obj = import_name.rsplit('.', 1)
        return getattr(__import__(module, None, None, [obj]), obj)
    else:
        return __import__(import_name)


def load_module(fullname):
    """Load thrift_file by fullname, fullname should have '_thrift' as
    suffix.

    The loader will replace the '_thrift' with '.thrift' and use it as
    filename to locate the real thrift file.
    """
    if not fullname.endswith("_thrift"):
        raise ImportError(
            "ThriftPy can only load module with '_thrift' suffix")

    if fullname in sys.modules:
        return sys.modules[fullname]

    if '.' in fullname:
        module_name, thrift_module_name = fullname.rsplit('.', 1)
        module = _import_module(module_name)
        path_prefix = os.path.dirname(os.path.abspath(module.__file__))
        path = os.path.join(path_prefix, thrift_module_name)
    else:
        path = fullname
    thrift_file = "{0}.thrift".format(path[:-7])

    module = load(thrift_file, module_name=fullname)
    sys.modules[fullname] = module
    return sys.modules[fullname]



# -*- coding: utf-8 -*-

from __future__ import absolute_import


class ThriftSyntaxError(SyntaxError):
    pass


class ThriftLexerError(ThriftSyntaxError):
    pass


class ThriftGrammerError(ThriftSyntaxError):
    pass



# -*- coding: utf-8 -*-

"""
    thriftpy._compat
    ~~~~~~~~~~~~~

    py2/py3 compatibility support.
"""

from __future__ import absolute_import

import sys
PY3 = sys.version_info[0] == 3

import types


if PY3:
    text_type = str
    string_types = (str,)

    def u(s):
        return s
else:
    text_type = unicode  # noqa
    string_types = (str, unicode)  # noqa

    def u(s):
        if not isinstance(s, text_type):
            s = s.decode("utf-8")
        return s


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass for py2 & py3

    This code snippet is copied from six."""
    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instantiation that replaces
    # itself with the actual metaclass.  Because of internal type checks
    # we also need to make sure that we downgrade the custom metaclass
    # for one level to something closer to type (that's why __call__ and
    # __init__ comes back from type etc.).
    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})


def init_func_generator(spec):
    """Generate `__init__` function based on TPayload.default_spec

    For example::

        spec = [('name', 'Alice'), ('number', None)]

    will generate::

        def __init__(self, name='Alice', number=None):
            kwargs = locals()
            kwargs.pop('self')
            self.__dict__.update(kwargs)

    TODO: The `locals()` part may need refine.
    """
    if not spec:
        def __init__(self):
            pass
        return __init__

    varnames, defaults = zip(*spec)
    varnames = ('self', ) + varnames

    def init(self):
        kwargs = locals()
        kwargs.pop('self')
        self.__dict__.update(kwargs)

    code = init.__code__
    if PY3:
        new_code = types.CodeType(len(varnames),
                                  0,
                                  len(varnames),
                                  code.co_stacksize,
                                  code.co_flags,
                                  code.co_code,
                                  code.co_consts,
                                  code.co_names,
                                  varnames,
                                  code.co_filename,
                                  "__init__",
                                  code.co_firstlineno,
                                  code.co_lnotab,
                                  code.co_freevars,
                                  code.co_cellvars)
    else:
        new_code = types.CodeType(len(varnames),
                                  len(varnames),
                                  code.co_stacksize,
                                  code.co_flags,
                                  code.co_code,
                                  code.co_consts,
                                  code.co_names,
                                  varnames,
                                  code.co_filename,
                                  "__init__",
                                  code.co_firstlineno,
                                  code.co_lnotab,
                                  code.co_freevars,
                                  code.co_cellvars)

    return types.FunctionType(new_code,
                              {"__builtins__": __builtins__},
                              argdefs=defaults)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import contextlib

from thriftpy.protocol import TBinaryProtocolFactory
from thriftpy.server import TThreadedServer
from thriftpy.thrift import TProcessor, TClient
from thriftpy.transport import (
    TBufferedTransportFactory,
    TServerSocket,
    TSocket,
)


def make_client(service, host, port,
                proto_factory=TBinaryProtocolFactory(),
                trans_factory=TBufferedTransportFactory(),
                timeout=None):
    socket = TSocket(host, port)
    if timeout:
        socket.set_timeout(timeout)
    transport = trans_factory.get_transport(socket)
    protocol = proto_factory.get_protocol(transport)
    transport.open()
    return TClient(service, protocol)


def make_server(service, handler, host, port,
                proto_factory=TBinaryProtocolFactory()):
    processor = TProcessor(service, handler)
    server_socket = TServerSocket(host=host, port=port)
    server = TThreadedServer(processor, server_socket,
                             iprot_factory=proto_factory)
    return server


@contextlib.contextmanager
def client_context(service, host, port,
                   proto_factory=TBinaryProtocolFactory(),
                   trans_factory=TBufferedTransportFactory(),
                   timeout=None):
    try:
        socket = TSocket(host, port)
        if timeout:
            socket.set_timeout(timeout)
        transport = trans_factory.get_transport(socket)
        protocol = proto_factory.get_protocol(transport)
        transport.open()
        yield TClient(service, protocol)
    finally:
        transport.close()



# -*- coding: utf-8 -*-

from __future__ import absolute_import

__all__ = ['TBinaryProtocol', 'TBinaryProtocolFactory',
           'TCyBinaryProtocol', 'TCyBinaryProtocolFactory',
           'TJSONProtocol', 'TJSONProtocolFactory']


from .binary import TBinaryProtocol, TBinaryProtocolFactory
from .cybin import TCyBinaryProtocol, TCyBinaryProtocolFactory
from .json import TJSONProtocol, TJSONProtocolFactory



# -*- coding: utf-8 -*-

"""
    thriftpy.thrift
    ~~~~~~~~~~~~~~~~~~

    Thrift simplified.
"""

from __future__ import absolute_import

import functools

from ._compat import init_func_generator, with_metaclass


def args2kwargs(thrift_spec, *args):
    arg_names = [item[1][1] for item in sorted(thrift_spec.items())]
    return dict(zip(arg_names, args))


class TType(object):
    STOP = 0
    VOID = 1
    BOOL = 2
    BYTE = 3
    I08 = 3
    DOUBLE = 4
    I16 = 6
    I32 = 8
    I64 = 10
    STRING = 11
    UTF7 = 11
    BINARY = 11  # This here just for parsing. For all purposes, it's a string
    STRUCT = 12
    MAP = 13
    SET = 14
    LIST = 15
    UTF8 = 16
    UTF16 = 17

    _VALUES_TO_NAMES = {
        STOP: 'STOP',
        VOID: 'VOID',
        BOOL: 'BOOL',
        BYTE: 'BYTE',
        I08: 'BYTE',
        DOUBLE: 'DOUBLE',
        I16: 'I16',
        I32: 'I32',
        I64: 'I64',
        STRING: 'STRING',
        UTF7: 'STRING',
        BINARY: 'STRING',
        STRUCT: 'STRUCT',
        MAP: 'MAP',
        SET: 'SET',
        LIST: 'LIST',
        UTF8: 'UTF8',
        UTF16: 'UTF16'
    }


class TMessageType(object):
    CALL = 1
    REPLY = 2
    EXCEPTION = 3
    ONEWAY = 4


class TPayloadMeta(type):
    def __new__(cls, name, bases, attrs):
        if "default_spec" in attrs:
            attrs["__init__"] = init_func_generator(attrs.pop("default_spec"))
        return super(TPayloadMeta, cls).__new__(cls, name, bases, attrs)


def gen_init(cls, thrift_spec=None, default_spec=None):
    if thrift_spec is not None:
        cls.thrift_spec = thrift_spec

    if "default_spec" is not None:
        cls.__init__ = init_func_generator(default_spec)
    return cls


class TPayload(with_metaclass(TPayloadMeta, object)):

    def read(self, iprot):
        iprot.read_struct(self)

    def write(self, oprot):
        oprot.write_struct(self)

    def __repr__(self):
        l = ['%s=%r' % (key, value) for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(l))

    def __str__(self):
        return repr(self)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
            self.__dict__ == other.__dict__

    def __ne__(self, other):
        return self != other


class TClient(object):
    def __init__(self, service, iprot, oprot=None):
        self._service = service
        self._iprot = self._oprot = iprot
        if oprot is not None:
            self._oprot = oprot
        self._seqid = 0

    def __getattr__(self, api):
        if api in self._service.thrift_services:
            return functools.partial(self._req, api)

        raise AttributeError("{} instance has no attribute '{}'".format(
            self.__class__.__name__, api))

    def __dir__(self):
        return self._service.thrift_services

    def _req(self, api, *args, **kwargs):
        _kw = args2kwargs(getattr(self._service, api + "_args").thrift_spec,
                          *args)
        kwargs.update(_kw)
        self._send(api, **kwargs)
        return self._recv(api)

    def _send(self, api, **kwargs):
        self._oprot.write_message_begin(api, TMessageType.CALL, self._seqid)
        args = getattr(self._service, api + "_args")()
        for k, v in kwargs.items():
            setattr(args, k, v)
        args.write(self._oprot)
        self._oprot.write_message_end()
        self._oprot.trans.flush()

    def _recv(self, api):
        fname, mtype, rseqid = self._iprot.read_message_begin()
        if mtype == TMessageType.EXCEPTION:
            x = TApplicationException()
            x.read(self._iprot)
            self._iprot.read_message_end()
            raise x
        result = getattr(self._service, api + "_result")()
        result.read(self._iprot)
        self._iprot.read_message_end()

        if hasattr(result, "success") and result.success is not None:
            return result.success

        # void api without throws
        if len(result.thrift_spec) == 0:
            return

        # check throws
        for k, v in result.__dict__.items():
            if k != "success" and v is not None:
                raise v

        # no throws & not void api
        if hasattr(result, "success"):
            raise TApplicationException(TApplicationException.MISSING_RESULT)


class TProcessor(object):
    """Base class for procsessor, which works on two streams."""

    def __init__(self, service, handler):
        self._service = service
        self._handler = handler

    def process_in(self, iprot):
        api, type, seqid = iprot.read_message_begin()
        if api not in self._service.thrift_services:
            iprot.skip(TType.STRUCT)
            iprot.read_message_end()
            return api, seqid, TApplicationException(TApplicationException.UNKNOWN_METHOD), None   # noqa
        else:
            args = getattr(self._service, api + "_args")()
            args.read(iprot)
            iprot.read_message_end()
            result = getattr(self._service, api + "_result")()

            # convert kwargs to args
            api_args = [args.thrift_spec[k][1]
                        for k in sorted(args.thrift_spec)]
            call = lambda: getattr(self._handler, api)(
                *(args.__dict__[k] for k in api_args)
            )
            return api, seqid, result, call

    def send_exception(self, oprot, api, exc, seqid):
        oprot.write_message_begin(api, TMessageType.EXCEPTION, seqid)
        exc.write(oprot)
        oprot.write_message_end()
        oprot.trans.flush()

    def send_result(self, oprot, api, result, seqid):
        oprot.write_message_begin(api, TMessageType.REPLY, seqid)
        result.write(oprot)
        oprot.write_message_end()
        oprot.trans.flush()

    def handle_exception(self, e, result):
        for k in sorted(result.thrift_spec):
            if result.thrift_spec[k][1] == "success":
                continue

            _, exc_name, exc_cls = result.thrift_spec[k]
            if isinstance(e, exc_cls):
                setattr(result, exc_name, e)
                break
        else:
            raise

    def process(self, iprot, oprot):
        api, seqid, result, call = self.process_in(iprot)
        if isinstance(result, TApplicationException):
            self.send_exception(oprot, api, result, seqid)
        else:
            try:
                result.success = call()
            except Exception as e:
                # raise if api don't have throws
                self.handle_exception(e, result)

            self.send_result(oprot, api, result, seqid)


class TException(TPayload, Exception):
    """Base class for all thrift exceptions."""


class TApplicationException(TException):
    """Application level thrift exceptions."""

    thrift_spec = {
        1: (TType.STRING, 'message'),
        2: (TType.I32, 'type'),
    }

    UNKNOWN = 0
    UNKNOWN_METHOD = 1
    INVALID_MESSAGE_TYPE = 2
    WRONG_METHOD_NAME = 3
    BAD_SEQUENCE_ID = 4
    MISSING_RESULT = 5
    INTERNAL_ERROR = 6
    PROTOCOL_ERROR = 7

    def __init__(self, type=UNKNOWN, message=None):
        super(TApplicationException, self).__init__()
        self.type = type
        self.message = message

    def __str__(self):
        if self.message:
            return self.message

        if self.type == self.UNKNOWN_METHOD:
            return 'Unknown method'
        elif self.type == self.INVALID_MESSAGE_TYPE:
            return 'Invalid message type'
        elif self.type == self.WRONG_METHOD_NAME:
            return 'Wrong method name'
        elif self.type == self.BAD_SEQUENCE_ID:
            return 'Bad sequence ID'
        elif self.type == self.MISSING_RESULT:
            return 'Missing result'
        else:
            return 'Default (unknown) TApplicationException'



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re
from os.path import join, dirname

with open(join(dirname(__file__), 'thriftpy', '__init__.py'), 'r') as f:
    version = re.match(r".*__version__ = '(.*?)'", f.read(), re.S).group(1)

from setuptools import setup, find_packages
from setuptools.extension import Extension

install_requires = [
    "ply==3.4",
]

tornado_requires = [
    "tornado>=4.0,<5.0",
    "toro==0.6"
]

dev_requires = [
    "cython>=0.20.2",
    "flake8>=2.2.2",
    "pytest>=2.6.0",
    "sphinx-rtd-theme>=0.1.6",
    "sphinx>=1.2.2",
] + tornado_requires


try:
    from Cython.Distutils import build_ext
    cython = True
except ImportError:
    cython = False

cmdclass = {}
ext_modules = []

if cython:
    ext_modules.append(Extension("thriftpy.transport.cytransport",
                                 ["thriftpy/transport/cytransport.pyx"]))
    ext_modules.append(Extension("thriftpy.protocol.cybin",
                                 ["thriftpy/protocol/cybin/cybin.pyx"]))
    cmdclass["build_ext"] = build_ext
else:
    ext_modules.append(Extension("thriftpy.transport.cytransport",
                                 ["thriftpy/transport/cytransport.c"]))
    ext_modules.append(Extension("thriftpy.protocol.cybin",
                                 ["thriftpy/protocol/cybin/cybin.c"]))

setup(name="thriftpy",
      version=version,
      description="Pure python implemention of Apache Thrift.",
      keywords="thrift python thriftpy",
      author="Lx Yu",
      author_email="i@lxyu.net",
      packages=find_packages(exclude=['benchmark', 'docs', 'tests']),
      entry_points={},
      url="https://thriftpy.readthedocs.org/",
      license="MIT",
      zip_safe=False,
      long_description=open("README.rst").read(),
      install_requires=install_requires,
      tests_require=tornado_requires,
      extras_require={
          "dev": dev_requires,
          "tornado": tornado_requires
      },
      cmdclass=cmdclass,
      ext_modules=ext_modules,
      classifiers=[
          "Topic :: Software Development",
          "Development Status :: 3 - Alpha",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.3",
          "Programming Language :: Python :: 3.4",
          "Programming Language :: Python :: Implementation :: CPython",
          "Programming Language :: Python :: Implementation :: PyPy",
      ])
